'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var constants = require('./constants.js');
var helpers = require('./helpers.js');

let c = 0;

const add = (node, bindings) => {
  node.$meta.bindings.unshift(...bindings);
};

const resolveSquares = (str) => {
  let parts = str.split(/(\[[^\]]+\])/).filter((v) => v);
  return parts.reduce((a, part) => {
    let v = part.charAt(0) === '[' ? '.' + part.replace(/\./g, ':') : part;
    return a + v;
  }, '');
};

const convertIndex = (path, context) => {
  for (let { index, prop } of context) {
    if (path === index) return `#${prop}`;
  }
  return path;
};

const resolve = (path, context) => {
  path = convertIndex(path, context);

  let isIndex = path.charAt(0) === '#';

  if (isIndex) path = path.slice(1);

  path = resolveSquares(path);

  let i = context.length;
  while (i--) {
    let { valueIdentifier, prop } = context[i];

    path = path
      .split('.')
      .map((v) => {
        let m = v.charAt(0) === '[';

        if (m) v = v.slice(1, -1);

        v = v.trim();

        if (v === valueIdentifier) v = prop + (m ? ':*' : '.*');

        return m ? `[${v}]` : v;
      })
      .join('.');
  }

  return isIndex ? `#${path}` : path;
};

const getParts = (value, context) =>
  value
    .trim()
    .split(/({{[^{}]+}})/)
    .filter((v) => v)
    .map((v) => {
      let match = v.match(/{{([^{}]+)}}/);

      if (match) {
        let m = match[1].trim();
        let negated = m.charAt(0) === '!';

        if (negated) m = m.slice(1);

        let fn = parseEventHandler(m, context);

        return fn
          ? {
              type: 'function',
              negated,
              ...fn,
            }
          : {
              type: 'key',
              value: resolve(m, context),
              negated,
            };
      }

      return {
        type: 'value',
        value: v,
      };
    });

let subscribers;

function parseElementNode(node, context) {
  let attrs = node.attributes;
  let i = attrs.length;
  while (i--) {
    parseAttributeNode(attrs[i], node, context);
  }
  return context;
}

const parseEventHandler = (value, context) => {
  let m = value.match(/(?:(\w+) => )?([^\(]+)(?:\(([^\)]*)\))/);

  if (!m) return;

  let event = m[1],
    method = m[2],
    args = m[3] && m[3].trim();

  args = args
    ? args
        .split(',')
        .filter((v) => v)
        .map((v) => v.trim())
        .map((k) => resolve(k, context))
    : null;

  return {
    event,
    method,
    args,
  };
};

const parseAttributeNode = ({ name, value }, node, context) => {
  if (name.startsWith('on')) {
    node.removeAttribute(name);
    let eventName = name.split('on')[1];

    subscribers.add(eventName);

    let { event, method, args } = parseEventHandler(value, context);

    add(node, [
      {
        type: 'call',
        eventName: eventName,
        event,
        method,
        args,
      },
    ]);

    return;
  }

  if (
    name === 'name' &&
    value &&
    (node.nodeName === 'INPUT' || node.nodeName === 'SELECT' || node.nodeName === 'TEXTAREA')
  ) {
    let path = resolve(value, context);

    subscribers.add('input');

    let binding = {
      type: constants.INPUT,
      path,
    };

    add(node, [
      binding,
      {
        type: 'set',
        eventName: 'input',
        path,
      },
    ]);
  }

  if (helpers.hasMustache(value)) {
    node.removeAttribute(name);

    add(node, [
      {
        name,
        parts: getParts(value, context),
        type: constants.ATTRIBUTE,
        context: context.slice(),
      },
    ]);
  }
};

const parseTextNode = (value, node, context) => {
  if (!helpers.hasMustache(value)) return;

  add(node, [
    {
      childIndex: Array.from(node.parentNode.childNodes).findIndex((v) => v === node),
      parts: getParts(value, context),
      type: constants.TEXT,
      context: context.slice(),
    },
  ]);
};

function parseEach(node) {
  let each = node.getAttribute('each');
  let m = each && each.match(/(.+)\s+in\s+(.+)/);
  if (!m) return;
  let [_, left, right] = m;
  let parts = left.match(/\(([^\)]+)\)/);
  let [valueIdentifier, index] = (parts ? parts[1].split(',') : [left]).map((v) => v.trim());

  return {
    prop: right.trim(),
    valueIdentifier,
    index,
    key: node.getAttribute('key') || 'id',
  };
}

let listCount = 0;

const bind = (element, mountNode) => {
  subscribers = new Set();
  c = 0;

  let parse = () => {
    let stack = [];

    function dispatch(node) {
      node.$meta = {
        rootNode: mountNode,
        index: c++,
        bindings: [],
      };

      switch (node.nodeType) {
        case node.TEXT_NODE: {
          parseTextNode(node.nodeValue, node, stack);
          break;
        }
        case node.ELEMENT_NODE: {
          if (node.nodeName === 'TEMPLATE') {
            let each = parseEach(node);

            if (each) {
              stack.push(each);
              helpers.walk(node.content, dispatch);
              let { key, prop } = helpers.last(stack);
              let path = resolve(prop, stack);

              let binding = {
                uid: key,
                path,
              };

              node.$meta.listId = listCount;

              add(node, [
                {
                  ...binding,
                  type: constants.LIST,
                },
              ]);

              let listNodeBinding = {
                ...binding,
                type: constants.LIST_ITEM,
              };

              for (let child of node.content.children) {
                child.$meta.listId = listCount;
                add(child, [listNodeBinding]);
              }

              stack.pop();
              listCount++;
            }
          } else {
            parseElementNode(node, stack);
          }
        }
      }
    }

    helpers.walk(element, dispatch);

    return Array.from(subscribers);
  };

  return parse();
};

exports.bind = bind;
exports.resolve = resolve;
