'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var constants = require('./constants.js');
var helpers = require('./helpers.js');
var cloneNode = require('./cloneNode.js');
var compareKeyedLists = require('./compareKeyedLists.js');

const getListItems = (template) => {
  let node = template.nextSibling;

  let nodes = [];

  while (node && node.$meta?.listId === template.$meta.listId) {
    nodes[node.$meta.blockIndex] = nodes[node.$meta.blockIndex] || [];
    nodes[node.$meta.blockIndex].push(node);
    node = node.nextSibling;
  }

  return nodes;
};

const updateList = (template, delta) => {
  let itemNodes = Array.from(template.content.children);
  let listItems = getListItems(template);
  let fragment = document.createDocumentFragment();

  listItems.forEach(helpers.removeNodes);

  delta.forEach((i, newIndex) => {
    let nodes = i === -1 ? itemNodes.map(cloneNode.cloneNode) : listItems[i];

    nodes.forEach((el) => {
      el.$meta.blockIndex = newIndex;
      fragment.appendChild(el);
    });
  });
  template.after(fragment);
};

let getPreviousValue = (node, binding) =>
  binding.type === constants.ATTRIBUTE
    ? helpers.attributeToProp(binding.name, node.getAttribute(binding.name)).value
    : node.textContent;

const getValue = (part, ctx, target) => {
  let { value, negated } = part;

  if (value.charAt(0) === '#') {
    return ctx[value.slice(1)];
  }

  let v = helpers.getValueAtPath(helpers.resolve(value, ctx, target), target);

  return negated ? !v : v;
};

const parseStyles = (value) => {
  let type = typeof value;

  if (type === 'string')
    return value.split(';').reduce((o, value) => {
      const [k, v] = value.split(':').map((v) => v.trim());
      if (k) o[k] = v;
      return o;
    }, {});

  if (type === 'object') return value;

  return {};
};

const joinStyles = (value) =>
  Object.entries(value)
    .map(([k, v]) => `${k}: ${v};`)
    .join(' ');

const convertStyles = (o) =>
  Object.keys(o).reduce((a, k) => {
    a[helpers.pascalToKebab(k)] = o[k];
    return a;
  }, {});

const applyComplexAttribute = (node, name, value) => {
  if (name === 'style') {
    value = joinStyles(
      convertStyles({
        ...parseStyles(node.getAttribute('style')),
        ...parseStyles(value),
      })
    );
  } else if (name === 'class') {
    switch (helpers.typeOf(value)) {
      case 'Array':
        value = value.join(' ');
        break;
      case 'Object':
        value = Object.keys(value)
          .reduce((a, k) => {
            if (value[k]) a.push(k);
            return a;
          }, [])
          .join(' ');
        break;
    }
  } else if (!helpers.isPrimitive(value)) {
    return (node[helpers.kebabToPascal(name)] = value);
  }

  helpers.applyAttribute(node, name, value);
};

const updateNode = (node, binding, newValue) =>
  binding.type === constants.ATTRIBUTE
    ? applyComplexAttribute(node, binding.name, newValue)
    : (node.textContent = newValue);

const updateBinding = (binding, node, ctx, p, viewmodel) => {
  if (binding.eventName) {
    binding.ctx = helpers.copy(ctx);
    binding.path && (binding.realPath = helpers.resolve(binding.path, ctx)); // @delete
    return;
  }

  if (binding.type === constants.LIST_ITEM) return (ctx[binding.path] = node.$meta.blockIndex);

  let oldValue = getPreviousValue(node, binding);

  if (binding.path) {
    const { path } = binding;
    const newValue = getValue({ value: path }, ctx, p);

    if (binding.type === constants.LIST) {
      const delta = compareKeyedLists.compareKeyedLists(binding.uid, node.$meta.blockData, newValue);
      node.$meta.blockData = newValue;
      return delta && updateList(node, delta);
    }

    if (oldValue === newValue) return;

    if (binding.type === constants.INPUT) {
      if (node.hasAttribute('multiple') && node.nodeName === 'SELECT') {
        Array.from(node.querySelectorAll('option')).forEach((option) => {
          option.selected = newValue.includes(option.value);
        });
        return;
      }

      switch (node.getAttribute('type')) {
        case 'checkbox':
          node.checked = newValue;
          if (node.checked) {
            node.setAttribute('checked', '');
          } else {
            node.removeAttribute('checked');
          }
          break;
        case 'radio':
          node.checked = newValue === node.getAttribute('value');
          if (node.checked) {
            node.setAttribute('checked', '');
          } else {
            node.removeAttribute('checked');
          }
          break;
        default:
          node.setAttribute('value', (node.value = newValue || ''));
          break;
      }
      return;
    }
  }

  const newValue = binding.parts.reduce((a, part) => {
    let { type, value } = part;

    let v = value;

    if (type === 'key') {
      v = getValue(part, ctx, p);
    } else if (type === 'function') {
      let args = part.args.map((value) => getValue({ value }, ctx, p));
      v = helpers.callFunctionAtPath(part.method, viewmodel, args);
    }

    return a ? a + v : v;
  }, '');

  if (newValue === oldValue) return;

  updateNode(node, binding, newValue);
};

let prev;

const updater = (mountNode, viewmodel, updatedCallback = () => {}) => (rootNode) => {
  let ctx = {};
  let p = helpers.copy(viewmodel);

  helpers.walk(rootNode, (node) => {
    if (node.$meta?.rootNode !== mountNode) return;

    let bindings = node.$meta.bindings;

    if (bindings) {
      bindings.forEach((binding) => updateBinding(binding, node, ctx, p, viewmodel));
    }
  });

  if (prev) updatedCallback(prev);

  prev = p;
};

exports.updater = updater;
